# Generated by Django 2.1.5 on 2020-10-07 23:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oauth2', '0002_fix_scope_size'),
    ]

    operations = [
        migrations.AddField(
            model_name='client',
            name='allow_public_token',
            field=models.BooleanField(blank=True, default=False, help_text='Allow public client tokens with only client_id and code'),
        ),
        migrations.AddField(
            model_name='client',
            name='authorize_every_time',
            field=models.BooleanField(blank=True, default=False),
        ),
    ]
